# -*- coding: utf-8 -*-
"""
Entry point for Django tests.

This script will setup the basic configuration needed by Django.
"""

import sys
from os.path import abspath, dirname, join

try:
    from django.conf import settings
    from django.test.utils import get_runner

    settings.configure(
        DEBUG=True,
        USE_TZ=True,
        DATABASES={
            'default': {
                'NAME': 'default',
                'ENGINE': 'django.db.backends.sqlite3',
            }
        },
        INSTALLED_APPS=[
            'django.contrib.auth',
            'django.contrib.contenttypes',
            'django.contrib.sites',
            'cerberus_ac',
        ],
        SITE_ID=1,
        MIDDLEWARE_CLASSES=(),
        CERBERUS_DEFAULT_RESPONSE=False,
        CERBERUS_SKIP_IMPLICIT=False,
        CERBERUS_LOG_ACCESS=True,
        CERBERUS_LOG_PRIVILEGES=True,
        CERBERUS_LOG_HIERARCHY=True,
        CERBERUS_MAPPING=(
            ('cerberus_ac.models.Role',
             {'name': 'role', 'attr': 'role'}),
            ('test_cerberus_ac.FakeUser',
             {'name': 'FakeUser', 'attr': 'user resource'}),
            ('test_cerberus_ac.FakeGroup',
             {'name': 'FakeGroup', 'attr': 'group resource'}),
            ('test_cerberus_ac.FakeResource',
             {'name': 'FakeResource', 'attr': 'resource'}),
        )
    )

    try:
        import django
        sys.path.append(abspath(join(dirname(__file__), 'src')))

        setup = django.setup
    except AttributeError:
        pass
    else:
        setup()

except ImportError:
    import traceback
    traceback.print_exc()
    raise ImportError('To fix this error, maybe run '
                      '`pip install -r requirements/test.txt`')


def run_tests(*test_args):
    """Discover and run tests."""
    if not test_args:
        test_args = ['tests']

    # Run tests
    runner = get_runner(settings)
    test_runner = runner()

    failures = test_runner.run_tests(test_args)

    if failures:
        sys.exit(bool(failures))


if __name__ == '__main__':
    run_tests(*sys.argv[1:])



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-04-07 18:06
from __future__ import unicode_literals

import cerberus_ac.models
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='AccessHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('role_type', models.CharField(blank=True, max_length=255, verbose_name='Role type')),
                ('role_id', models.CharField(blank=True, max_length=255, verbose_name='Role ID')),
                ('response', models.NullBooleanField(default=None, verbose_name='Response')),
                ('response_type', models.CharField(choices=[('d', 'by default'), ('i', 'implicitly'), ('e', 'explicitly')], max_length=1, verbose_name='Response type')),
                ('access_type', models.CharField(max_length=255, verbose_name='Access')),
                ('resource_type', models.CharField(max_length=255, verbose_name='Resource type')),
                ('resource_id', models.CharField(blank=True, max_length=255, verbose_name='Resource ID')),
                ('datetime', models.DateTimeField(default=django.utils.timezone.now, verbose_name='Date and time')),
                ('conveyor_type', models.CharField(blank=True, max_length=255, verbose_name='Conveyor type')),
                ('conveyor_id', models.CharField(blank=True, max_length=255, verbose_name='Conveyor ID')),
            ],
            options={
                'verbose_name_plural': 'Access history',
                'verbose_name': 'Access history',
            },
        ),
        migrations.CreateModel(
            name='PrivilegeHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('privilege_id', models.PositiveIntegerField(null=True, verbose_name='Privilege reference ID')),
                ('action', models.CharField(choices=[('c', 'create'), ('u', 'update'), ('d', 'delete')], max_length=1, verbose_name='Action')),
                ('datetime', models.DateTimeField(auto_now_add=True, verbose_name='Date and time')),
                ('role_type', models.CharField(blank=True, max_length=255, verbose_name='Role type')),
                ('role_id', models.CharField(blank=True, max_length=255, verbose_name='Role ID')),
                ('authorized', models.NullBooleanField(default=None, verbose_name='Authorized')),
                ('access_type', models.CharField(blank=True, max_length=255, verbose_name='Access type')),
                ('resource_type', models.CharField(blank=True, max_length=255, verbose_name='Resource type')),
                ('resource_id', models.CharField(blank=True, max_length=255, verbose_name='Resource ID')),
            ],
            options={
                'verbose_name_plural': 'Privilege history',
                'verbose_name': 'Privilege history',
            },
        ),
        migrations.CreateModel(
            name='Role',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(max_length=255)),
                ('rid', models.CharField(blank=True, max_length=255)),
            ],
            options={
                'verbose_name_plural': 'Roles',
                'verbose_name': 'Role',
            },
            bases=(models.Model, cerberus_ac.models.RoleMixin),
        ),
        migrations.CreateModel(
            name='RoleHierarchy',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('role_type_a', models.CharField(max_length=255)),
                ('role_id_a', models.CharField(blank=True, max_length=255)),
                ('role_type_b', models.CharField(max_length=255)),
                ('role_id_b', models.CharField(blank=True, max_length=255)),
            ],
            options={
                'verbose_name_plural': 'Role hierarchy',
                'verbose_name': 'Role hierarchy',
            },
        ),
        migrations.CreateModel(
            name='RolePrivilege',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('role_type', models.CharField(max_length=255, verbose_name='Role type')),
                ('role_id', models.CharField(blank=True, max_length=255, verbose_name='Role ID')),
                ('authorized', models.BooleanField(default=False, verbose_name='Authorized')),
                ('access_type', models.CharField(max_length=255, verbose_name='Access type')),
                ('resource_type', models.CharField(max_length=255, verbose_name='Resource type')),
                ('resource_id', models.CharField(blank=True, max_length=255, verbose_name='Resource ID')),
                ('creation_date', models.DateTimeField(auto_now_add=True, verbose_name='Created')),
                ('modification_date', models.DateTimeField(auto_now=True, verbose_name='Last modified')),
            ],
            options={
                'verbose_name_plural': 'Role privileges',
                'verbose_name': 'Role privilege',
            },
        ),
        migrations.AlterUniqueTogether(
            name='roleprivilege',
            unique_together=set([('role_type', 'role_id', 'access_type', 'resource_type', 'resource_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='rolehierarchy',
            unique_together=set([('role_type_a', 'role_id_a', 'role_type_b', 'role_id_b')]),
        ),
        migrations.AlterUniqueTogether(
            name='role',
            unique_together=set([('type', 'rid')]),
        ),
        migrations.AddField(
            model_name='privilegehistory',
            name='reference',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='history', to='cerberus_ac.RolePrivilege', verbose_name='Privilege reference'),
        ),
        migrations.AddField(
            model_name='privilegehistory',
            name='user',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='privileges_changes', to=settings.AUTH_USER_MODEL, verbose_name='User'),
        ),
    ]



# -*- coding: utf-8 -*-

"""App module providing the application settings class."""

import importlib

from django.apps import AppConfig

import appsettings as aps


class CerberusACConfig(AppConfig):
    name = 'cerberus_ac'
    verbose_name = 'Cerberus AC'

    def ready(self):
        AppSettings.check()


def _import(complete_path):
    module_name = '.'.join(complete_path.split('.')[:-1])
    imported_module = importlib.import_module(name=module_name)
    function_or_class = getattr(imported_module, complete_path.split('.')[-1])
    return function_or_class


class Mapping(object):
    """Mapping class to map roles/resources names to their classes."""

    def __init__(self, mapping):
        """
        Initialization method.

        Args:
            mapping (dict): CERBERUS_MAPPING setting.
        """
        self.mapping = mapping

    def get_class(self, name):
        """
        Return the class given the name of a role/resource.

        Args:
            name (str): the type of role/resource.

        Returns:
            class: the corresponding the role/resource class.
        """
        for k, v in self.mapping:
            if v['name'] == name:
                return _import(k)
        return None

    def get_instance(self, name, id):
        """
        Return an instance given a role/resource type and an ID.

        Args:
            name (str): the type of role/resource.
            id (int): an integer or None.

        Returns:
            obj: the instance or a (name, id) tuple if not found.
        """
        cls = self.get_class(name)
        if cls:
            if hasattr(cls, 'objects') and id:
                try:
                    return cls.objects.get(id=id)
                except (cls.DoesNotExist, ValueError):
                    return None
            return None
        from .models import Role
        try:
            return Role.objects.get(type=name, rid=id)
        except Role.DoesNotExist:
            return None

    def get_type(self, obj):
        """
        Return the type of a role/resource given a Python object.

        Args:
            obj (obj): a Python object (class are accepted).

        Returns:
            str: the role/resource type.
        """
        if isinstance(obj, type):
            class_name = obj.__name__
        else:
            class_name = obj.__class__.__name__
        for k, v in self.mapping:
            # FIXME: use complete path, not just the end
            if k.split('.')[-1] == class_name:
                return v['name']
        return class_name

    def user_classes(self):
        """Return the user-role classes."""
        return [_import(k) for k, v in self.mapping
                if 'user' in v['attr'].split()]

    def group_classes(self):
        """Return the group-role classes."""
        return [_import(k) for k, v in self.mapping
                if 'group' in v['attr'].split()]

    def role_types(self):
        """Return the role types."""
        return [v['name'] for k, v in self.mapping
                if len({'user', 'group', 'role'} &
                       set(v['attr'].split())) > 0]

    def role_classes(self):
        """Return the role classes."""
        return [_import(k) for k, v in self.mapping
                if len({'user', 'group', 'role'} &
                       set(v['attr'].split())) > 0]

    def resource_types(self):
        """Return the resource types."""
        return [v['name'] for k, v in self.mapping
                if 'resource' in v['attr'].split()]

    def resource_classes(self):
        """Return the resource classes."""
        return [_import(k) for k, v in self.mapping
                if 'resource' in v['attr'].split()]


class MappingSetting(aps.Setting):
    def checker(self, name, value):
        """Check the value of given mapping setting."""
        if not isinstance(value, tuple):
            raise ValueError('%s must be a tuple' % name)
        if not all(isinstance(o, tuple) for o in value):
            raise ValueError(
                '%s must be a tuple of (key, value) tuples' % name)
        for k, v in value:
            if not isinstance(k, str):
                raise ValueError('Keys in %s must be str' % name)
            if not isinstance(v, dict):
                raise ValueError('Values in %s must be dict' % name)
            if set(v.keys()) != {'name', 'attr'}:
                raise ValueError('Values in %s must be dict '
                                 'with name and attr keys' % name)
        _ = [o[1] for o in value]
        if {x['name'] for x in _ if _.count(x['name']) > 1}:
            raise ValueError(
                'Names in %s values must be unique' % name)

    def transform(self, value):
        return Mapping(value)


class AppSettings(aps.AppSettings):
    """
    Application settings class.

    Settings:
    - default_response (bool):
    - skip_implicit (bool):
    - log_access (bool):
    - log_privileges (bool):
    - log_hierarchy (bool):
    - mapping (tuple):
    - namespace (str):
    """

    allow_update_own_privileges = aps.BooleanSetting(default=False)
    default_response = aps.BooleanSetting(default=False)
    skip_implicit = aps.BooleanSetting(default=False)
    log_access = aps.BooleanSetting(default=True)
    log_privileges = aps.BooleanSetting(default=True)
    log_hierarchy = aps.BooleanSetting(default=True)
    namespace = aps.StringSetting(default='')
    mapping = MappingSetting(default=())
    access_permission = aps.StringSetting(default='read')

    class Meta:
        setting_prefix = 'CERBERUS_'



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-05-15 09:13
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cerberus_ac', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='privilegehistory',
            name='authorized',
            field=models.NullBooleanField(default=None, verbose_name='Authorized'),
        ),
    ]



# -*- coding: utf-8 -*-

"""URLs module."""

from django.conf.urls import url

from . import views
from .apps import AppSettings


def cerberus_urlpatterns(admin_view_func=lambda x: x):
    """
    Function returning URLs, optionally wrapped into an admin_view function.

    Args:
        admin_view_func (func):
            the ``admin_view`` function of an ``AdminSite``. Will default to
            identity.

    Returns:
        list: list of URLs patterns.
    """
    return [
        url(r'^$', admin_view_func(views.Index.as_view()),
            name='index'),

        url(r'^privileges/$',
            admin_view_func(views.Privileges.as_view()),
            name='privileges'),

        # role views
        url(r'^privileges/view/(?P<role_type>\w+)/(?P<resource_type>\w+)/$',
            admin_view_func(views.ViewPrivileges.as_view()),
            name='view_privileges'),
        url(r'^privileges/edit/(?P<role_type>\w+)/(?P<resource_type>\w+)/$',
            admin_view_func(views.EditPrivileges.as_view()),
            name='edit_privileges'),
        url(r'^privileges/edit/(?P<role_type>\w+)/(?P<resource_type>\w+)/'
            r'(?P<role_id>\w+)/(?P<resource_id>\w+)/(?P<privilege>\w+)/'
            r'(?P<action>\w+)/$',
            admin_view_func(views.ajax_edit_privileges),
            name='ajax_edit_privileges'),
        url(r'^privileges/load_rr/(?P<role_type>\w+)/(?P<resource_type>\w+)/',
            admin_view_func(views.ajax_load_roles_and_resources),
            name='ajax_load_roles_and_resources'),
        url(r'^privileges/load/(?P<role_type>\w+)/(?P<role_id>\w+)/'
            r'(?P<resource_type>\w+)/',
            admin_view_func(views.ajax_load_privileges),
            name='ajax_load_privileges'),

        # Member Info
        url(r'^members/$',
            admin_view_func(views.MemberList.as_view()),
            name='member_list'),
        url(r'^members/(?P<member_id>\d+)/$',
            admin_view_func(views.MemberInfo.as_view()),
            name='member_info'),

        # hierarchy
        url(r'^role/hierarchy/$',
            admin_view_func(views.ViewRoleHierarchy.as_view()),
            name='role_hierarchy')
    ]


app_name = AppSettings.namespace.get_value()
urlpatterns = cerberus_urlpatterns()



# -*- coding: utf-8 -*-

"""Admin module."""

from django.contrib import admin
from django.core.urlresolvers import reverse
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _

from .apps import AppSettings
from .models import (
    AccessHistory, PrivilegeHistory, Role, RoleHierarchy, RolePrivilege)

app_settings = AppSettings()


# class SecurityAdmin(AdminSite):
#     pass
#
#
# class DataAdmin(AdminSite):
#     pass
#
#
# class AuditAdmin(AdminSite):
#     pass
#
#
# security_admin_site = SecurityAdmin(name='SecurityAdmin')
# data_admin_site = DataAdmin(name='DataAdmin')
# audit_admin_site = AuditAdmin(name='AuditAdmin')
#
# Use decorator like @security_admin_site.register(AccessHistory)


def obj_link_generator(type_attr, id_attr, short_description):
    def obj_link(obj):
        obj_type = getattr(obj, type_attr)
        obj_id = getattr(obj, id_attr)
        instance = app_settings.mapping.get_instance(
            obj_type, obj_id)
        if instance is None:
            return '-'
        info = (instance._meta.app_label, instance._meta.model_name)
        admin_url = reverse('admin:%s_%s_change' % info,
                            args=(instance.pk,))
        return format_html('<a href="{}">{}</a>', admin_url, instance)
    obj_link.short_description = short_description
    return obj_link


role_link = obj_link_generator(
    'role_type', 'role_id', _('Role link'))
resource_link = obj_link_generator(
    'resource_type', 'resource_id', _('Resource link'))
hierarchy_role_link_a = obj_link_generator(
    'role_type_a', 'role_id_a', _('Role link A'))
hierarchy_role_link_b = obj_link_generator(
    'role_type_b', 'role_id_b', _('Role link B'))


class RoleAdmin(admin.ModelAdmin):
    """Role admin class."""

    list_display = ('type', 'rid')


class RolePrivilegeAdmin(admin.ModelAdmin):
    """Role privilege admin class."""

    actions_on_top = True
    actions_on_bottom = True

    list_display = (
        '__str__',
        'role_type',
        'role_id',
        role_link,
        'authorized',
        'access_type',
        'resource_type',
        'resource_id',
        resource_link,
        'creation_date',
        'modification_date')

    list_filter = (
        'role_type', 'resource_type', 'authorized', 'access_type',
        'modification_date')
    date_hierarchy = 'creation_date'

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        record = PrivilegeHistory(
            user=request.user, action={False: PrivilegeHistory.CREATE}.get(
                change, PrivilegeHistory.UPDATE))
        record.update_from_privilege(obj)


class RoleHierarchyAdmin(admin.ModelAdmin):
    """Role hierarchy admin class."""

    list_display = (
        '__str__',
        'role_type_a',
        'role_id_a',
        hierarchy_role_link_a,
        'role_type_b',
        'role_id_b',
        hierarchy_role_link_b)


class AccessHistoryAdmin(admin.ModelAdmin):
    """Acces history admin class."""

    list_display = (
        'datetime',
        'role_type',
        'role_id',
        role_link,
        'response',
        'response_type',
        'access_type',
        'resource_type',
        resource_link,
        'resource_id',
        'conveyor_type',
        'conveyor_id')


class PrivilegeHistoryAdmin(admin.ModelAdmin):
    """Privilege history admin class."""

    list_display = (
        'datetime',
        'user',
        'action',
        'role_type',
        'role_id',
        role_link,
        'authorized',
        'access_type',
        'resource_type',
        'resource_id',
        resource_link)


# class HierarchyHistoryAdmin(admin.ModelAdmin):
#     pass


admin.site.register(Role, RoleAdmin)
admin.site.register(RolePrivilege, RolePrivilegeAdmin)
admin.site.register(RoleHierarchy, RoleHierarchyAdmin)
admin.site.register(AccessHistory, AccessHistoryAdmin)
admin.site.register(PrivilegeHistory, PrivilegeHistoryAdmin)
# admin.site.register(HierarchyHistory, HierarchyHistoryAdmin)



# -*- coding: utf-8 -*-

"""Views module."""

import json

from django.http import HttpResponse
from django.utils.translation import ugettext as _
from django.views.defaults import bad_request

from suit_dashboard import Box, Column, DashboardView, Grid, Row

from .apps import AppSettings
from .models import RoleHierarchy, RolePrivilege
from .utils import get_role_id, get_role_type

app_settings = AppSettings()


class Index(DashboardView):
    """Cerberus menu."""

    title = _('Index - Cerberus AC')
    crumbs = (
        {'name': _('Home'), 'url': 'admin:index'},
        {'name': _('Cerberus AC'), 'url': 'admin:cerberus:index'},
    )
    grid = Grid(Row(Column(Box(
        title='Cerberus Access Control',
        template='cerberus_ac/index.html'))))


class MemberList(Index):
    """View to see the list of members."""

    title = _('Member List - Cerberus AC')
    crumbs = ({'name': _('Member List'), 'url': 'admin:cerberus:member_list'},)

    def get(self, request, *args, **kwargs):
        member_class = app_settings.mapping.user_classes()[0]
        members = member_class.objects.all()

        # members = get_paginated_data(
        #     members, request.GET.get('page_user_list'), 50)

        self.grid = Grid(Row(Column(Box(
            title=_('Member list'),
            template='cerberus_ac/member_list.html',
            context={'members': members}))))

        return super(MemberList, self).get(request, *args, **kwargs)


class MemberInfo(Index):
    """View to see member info."""

    title = _('Member Info - Cerberus AC')
    crumbs = (
        {'name': _('Member List'), 'url': 'admin:cerberus:member_list'},
        {'name': _('Member Info')}
    )

    def get(self, request, *args, **kwargs):
        user_class = app_settings.mapping.user_classes()[0]
        user = user_class.objects.get(id=kwargs['member_id'])

        self.grid = Grid(Row(Column(Box(
            template='cerberus_ac/member_info.html',
            context={'member': user}))))

        return super(MemberInfo, self).get(request, *args, **kwargs)


class Privileges(Index):
    """Menu for privileges."""

    context = {'role_types': [{
        'verbose': app_settings.mapping.get_class(t)._meta.verbose_name,
        'slug': t} for t in app_settings.mapping.role_types()
    ], 'resource_types': [{
        'verbose': app_settings.mapping.get_class(t)._meta.verbose_name,
        'slug': t} for t in app_settings.mapping.resource_types()
    ]}

    title = _('Privileges Matrix - Cerberus AC')
    crumbs = ({'name': _('Privileges Matrix'),
               'url': 'admin:cerberus:privileges'},)
    grid = Grid(Row(Column(Box(
        title=_('Privileges Matrix'),
        template='cerberus_ac/privileges.html',
        context=context))))


class ViewPrivileges(Privileges):
    """View to see user privileges."""

    title = _('View Privileges - Cerberus AC')
    crumbs = ({'name': _('View'), 'url': 'admin:cerberus:view_privileges'}, )

    def get(self, request, *args, **kwargs):
        role_type = kwargs.pop('role_type')
        resource_type = kwargs.pop('resource_type')
        role_class = app_settings.mapping.get_class(role_type)
        resource_class = app_settings.mapping.get_class(resource_type)
        role_instances = role_class.objects.all()
        resource_instances = resource_class.objects.all()
        self.grid = Grid(Row(Column(Box(
            template='cerberus_ac/view_privileges.html',
            context={'roles': role_instances, 'resources': resource_instances}
        ))))

        return super(ViewPrivileges, self).get(request, *args, **kwargs)


class EditPrivileges(Privileges):
    """View to edit user privileges."""

    title = _('Edit Privileges - Cerberus AC')
    crumbs = ({'name': _('Edit')}, )

    def get(self, request, *args, **kwargs):
        role_type = kwargs.pop('role_type')
        resource_type = kwargs.pop('resource_type')
        role_class = app_settings.mapping.get_class(role_type)
        resource_class = app_settings.mapping.get_class(resource_type)
        role_instances = role_class.objects.all()
        resource_instances = resource_class.objects.all()

        self.grid = Grid(Row(Column(
            Box(title=_('Edit role privileges between %s and %s') % (
                    role_class._meta.verbose_name_plural,
                    resource_class._meta.verbose_name_plural),
                template='cerberus_ac/edit_privileges.html',
                context={'roles': role_instances,
                         'resources': resource_instances,
                         'role_type': role_type,
                         'resource_type': resource_type})
        )))

        return super(EditPrivileges, self).get(request, *args, **kwargs)


class ViewRoleHierarchy(Index):
    """Role hierarchy view."""

    title = _('Role Hierarchy Graph - Cerberus AC')
    crumbs = ({'name': _('Role Hierarchy Graph'),
               'url': 'admin:cerberus:role_hierarchy'},)

    def get(self, request, *args, **kwargs):
        # link
        # {"source": 3, "target": 11}
        # node
        # {"size": 10, "score": 0, "id": "security", "type": "circle"}

        roles = []
        links = []
        count = {}
        for tuple4 in list(RoleHierarchy.objects.values_list(
                'role_type_a', 'role_id_a', 'role_type_b', 'role_id_b')):
            role_a = ('%s %s' % (tuple4[0], tuple4[1])) if tuple4[1] else tuple4[0]  # noqa
            role_b = (('%s %s' % (tuple4[2], tuple4[3])) if tuple4[3] else tuple4[2])  # noqa
            if role_a not in count:
                count[role_a] = 1
                roles.append(role_a)
            else:
                count[role_a] += 1
            if role_b not in count:
                count[role_b] = 1
                roles.append(role_b)
            else:
                count[role_b] += 1
            # I think we can add scores on links
            links.append({'source': roles.index(role_b),
                          'target': roles.index(role_a)})

        # possible types: circle, square, triangle-up/down, diamond, cross
        data = {
            'directed': True,
            'multigraph': False,
            'links': links,
            'nodes': [
                {'size': min(count[r], 50), 'score': min(count[r] / 100, 1),
                 'id': r if ' ' not in r else str(
                     app_settings.mapping.get_instance(*r.split(' '))),
                 'type': {
                     # FIXME: put this outside of cerberus: it's genida related
                     'member': 'circle',
                     'relative': 'square',
                     'cohort': 'diamond',
                     'gene': 'cross',
                     'copy_number_variant': 'cross',
                 }.get(r.split(' ')[0], 'triangle-down')}
                for r in roles
            ]
        }

        self.grid = Grid(Row(Column(Box(
            title='Role Hierarchy Graph',
            template='cerberus_ac/view_role_hierarchy.html',
            context=json.dumps(data)))))

        return super(ViewRoleHierarchy, self).get(request, *args, **kwargs)


# Ajax views ------------------------------------------------------------------
def ajax_edit_privileges(request,
                         role_type,
                         resource_type,
                         role_id,
                         resource_id,
                         privilege,
                         action):
    user = request.user
    success = False

    if not request.is_ajax():
        return bad_request(request, ValueError('not an ajax request'))
    elif not user.can('update', 'role_privilege'):
        message = _("You don't have the authorization to edit privileges")

    elif (not app_settings.allow_update_own_privileges and
          get_role_type(user) == role_type and
          get_role_id(user) == role_id):
        message = _("You can't edit your own privileges")

    elif action in ('allow', 'deny'):
        func = getattr(RolePrivilege, action)
        privilege, created = func(role_type, role_id, privilege,
                                  resource_type, resource_id, user)
        success = True
        if created:
            message = _('Successfully created privilege')
        else:
            message = _('Successfully updated privilege')
    elif action == 'forget':
        deleted = RolePrivilege.forget(role_type, role_id, privilege,
                                       resource_type, resource_id, user)
        if deleted:
            success = True
            message = _('Successfully deleted privilege')
        else:
            message = _('Privilege did not exist')
    else:
        message = _('Invalid request: action parameter accepts '
                    'allow, deny or forget')

    return HttpResponse(json.dumps({'success': success, 'message': message}),
                        content_type='application/json')


def ajax_load_roles_and_resources(request, role_type, resource_type):
    if not request.is_ajax():
        return bad_request(request, ValueError('not an ajax request'))

    role_class = app_settings.mapping.get_class(role_type)
    resource_class = app_settings.mapping.get_class(resource_type)

    role_instances = role_class.objects.order_by('id')
    resource_instances = resource_class.objects.order_by('id')

    role_instances = [(r.id, str(r)) for r in role_instances]
    resource_instances = [(r.id, str(r)) for r in resource_instances]

    data = {'roles': role_instances, 'resources': resource_instances}

    return HttpResponse(json.dumps(data), content_type="application/json")


def ajax_load_privileges(request, role_type, role_id, resource_type):
    if not request.is_ajax():
        return bad_request(request, ValueError('not an ajax request'))

    resource_class = app_settings.mapping.get_class(resource_type)
    resource_instances = resource_class.objects.order_by('id')

    default_access_types = ('read', 'update', 'delete')

    data = []

    for resource in resource_instances:
        current_privileges = {}
        db_access_types = []
        for privilege in RolePrivilege.objects.filter(
                role_type=role_type, role_id=role_id,
                resource_type=resource_type, resource_id=resource.id):
            current_privileges[privilege.access_type] = privilege.authorized
            db_access_types.append(privilege.access_type)
        for default_access_type in default_access_types:
            if default_access_type not in db_access_types:
                current_privileges[default_access_type] = None
        data.append(current_privileges)

    return HttpResponse(json.dumps(data), content_type="application/json")



# -*- coding: utf-8 -*-

"""
Models and mixins for access control.

- RoleMixin
- Role
- RoleHierarchy
- RolePrivilege
- PrivilegeHistory
- AccessHistory
"""

from django.conf import settings
from django.db import models
from django.db.models import QuerySet
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from .apps import AppSettings
from .utils import get_resource_type_and_id, get_role_type_and_id

app_settings = AppSettings()


class RoleMixin(object):
    """
    Mixin for Role models / classes.

    This mixin provides shortcuts methods to manipulate and check the
    privileges granted to a role, be it a custom class or Role instance.
    """

    def heirs(self, search=None):
        """Return the children of this role."""
        role_type, role_id = get_role_type_and_id(self)
        if search is not None:
            return [app_settings.mapping.get_instance(*h)
                    for h in RoleHierarchy.all_heirs(
                    role_type, role_id, search)]
        return [app_settings.mapping.get_instance(*h)
                for h in RoleHierarchy.heirs(role_type, role_id)]

    def has_role(self, role, role_id='', direct=False):
        """
        Check if this role inherits privileges from the given role.

        Args:
            role (str/role): a string describing the role, or a role instance.
            role_id (str): only used when role is a string.
            direct (bool): check only direct relations or transitive ones.

        Returns:
            bool: True or False
        """
        role_type_a, role_id_a = get_role_type_and_id(self)
        role_type_b, role_id_b = get_role_type_and_id(role, role_id)
        if direct:
            try:
                RoleHierarchy.objects.get(
                    role_type_a=role_type_a, role_id_a=role_id_a,
                    role_type_b=role_type_b, role_id_b=role_id_b)
                return True
            except RoleHierarchy.DoesNotExist:
                return False
        else:
            if (role_type_b, role_id_b) in RoleHierarchy.all_conveyors(
                    role_type_a, role_id_a):
                return True
            return False

    def take_role(self, role, role_id=''):
        """
        Set this role as child of the given role.

        The child role will then inherit privileges from the parent role.

        Args:
            role (str/role): a string describing the role, or a role instance.
            role_id (str): only used when role is a string.

        Returns:
            obj: the created RoleHierarchy object.
        """
        role_type_a, role_id_a = get_role_type_and_id(self)
        role_type_b, role_id_b = get_role_type_and_id(role, role_id)
        return RoleHierarchy.objects.create(
            role_type_a=role_type_a, role_id_a=role_id_a,
            role_type_b=role_type_b, role_id_b=role_id_b)

    def conveyors(self, search=None):
        """Return the roles conveying privileges to this role."""
        role_type, role_id = get_role_type_and_id(self)
        if search is not None:
            return [app_settings.mapping.get_instance(*c)
                    for c in RoleHierarchy.all_conveyors(
                    role_type, role_id, search)]
        return [app_settings.mapping.get_instance(*c)
                for c in RoleHierarchy.conveyors(role_type, role_id)]

    def conveys_role_to(self, role, role_id=''):
        """
        Check if this role conveys privileges to the given role.

        Args:
            role (str/role): a string describing the role, or a role instance.
            role_id (str): only used when role is a string.

        Returns:
            bool: True or False
        """
        role_type_a, role_id_a = get_role_type_and_id(role, role_id)
        role_type_b, role_id_b = get_role_type_and_id(self)
        try:
            RoleHierarchy.objects.get(
                role_type_a=role_type_a, role_id_a=role_id_a,
                role_type_b=role_type_b, role_id_b=role_id_b)
            return True
        except RoleHierarchy.DoesNotExist:
            return False

    def convey_role_to(self, role, role_id=''):
        """
        Set this role as parent of the given role.

        The parent role will then transmit privileges to the child role.

        Args:
            role (str/role): a string describing the role, or a role instance.
            role_id (str): only used when role is a string.

        Returns:
            obj: the created RoleHierarchy object.
        """
        role_type_a, role_id_a = get_role_type_and_id(role, role_id)
        role_type_b, role_id_b = get_role_type_and_id(self)
        return RoleHierarchy.objects.create(
            role_type_a=role_type_a, role_id_a=role_id_a,
            role_type_b=role_type_b, role_id_b=role_id_b)

    def has_access_to(self, resource, resource_id=''):
        return self.can(app_settings.access_permission, resource, resource_id,
                        log=False)

    def accessible(self, resources):
        if isinstance(resources, (str, QuerySet)):
            if isinstance(resources, str):
                resource_type = resources
                model = app_settings.mapping.get_class(resource_type)
                if not model:
                    return []
                queryset = model.objects
            else:
                model = resources.model
                resource_type = app_settings.mapping.get_type(model)
                queryset = resources

            self_type, self_id = get_role_type_and_id(self)
            roles = [(self_type, self_id)]
            roles.extend(RoleHierarchy.all_conveyors(self_type, self_id))
            resources_id = []
            for role_type, role_id in roles:
                resources_id.extend(list(RolePrivilege.objects.filter(
                    role_type=role_type, role_id=role_id,
                    resource_type=resource_type).values_list(
                    'resource_id', flat=True
                )))

            return queryset.filter(id__in=resources_id)
        else:
            # FIXME: can optimize a lot here by grouping objects by type,
            # and doing like above, once for each type
            return [r for r in resources if self.has_access_to(r)]

    def can(self, perm, resource, resource_id='',
            skip_implicit=None, log=None):
        """
        Check if this role has privilege ``perm`` on resource.

        Args:
            perm (str): string describing permission or privilege.
            resource (str/obj): a resource instance or a string describing it.
            resource_id (str): only used when resource is a string.
            skip_implicit (bool): True, False or None for project's default.
            log (bool): True, False or None for project's default.

        Returns:
            bool: True or False
        """
        role_type, role_id = get_role_type_and_id(self)
        resource_type, resource_id = get_resource_type_and_id(resource, resource_id)  # noqa
        return RolePrivilege.authorize(
            role_type=role_type, role_id=role_id, perm=perm,
            resource_type=resource_type, resource_id=resource_id,
            skip_implicit=skip_implicit, log=log)

    # def allow(perm, role, role_id=''): pass
    # def deny(perm, role, role_id=''): pass
    # def forget(perm, role, role_id=''): pass


# TODO: write the according Resource model
class ResourceMixin(object):
    """Mixin for Resource models / classes."""

    # TODO: implement methods
    # def accessors(role_type=None): return []
    # def is_accessed_by(role, role_id=''): return False

    # def allow(perm, role, role_id=''): pass
    # def deny(perm, role, role_id=''): pass
    # def forget(perm, role, role_id=''): pass

    # def accept(perm, role, role_id='', skip_implicit=None, log=None): pass


class Role(models.Model, RoleMixin):
    """Concrete model for roles."""

    type = models.CharField(max_length=255)
    rid = models.CharField(max_length=255, blank=True)

    class Meta:
        """Meta class for Django."""

        verbose_name = _('Role')
        verbose_name_plural = _('Roles')
        unique_together = ('type', 'rid')

    def __str__(self):
        if self.rid:
            return '%s %s' % (self.type, self.rid)
        return self.type

    def role_type(self):
        """Override role_type method and return type attribute."""
        return self.type

    def role_id(self):
        """Override role_id method and return rid attribute."""
        return self.rid


class RoleHierarchy(models.Model):
    """Role hierarchy model."""

    BREADTH_FIRST = 0
    DEPTH_FIRST = 1

    role_type_a = models.CharField(max_length=255)
    role_id_a = models.CharField(max_length=255, blank=True)
    role_type_b = models.CharField(max_length=255)
    role_id_b = models.CharField(max_length=255, blank=True)

    class Meta:
        """Meta class for Django."""

        verbose_name = _('Role hierarchy')
        verbose_name_plural = _('Role hierarchy')
        unique_together = (
            'role_type_a', 'role_id_a',
            'role_type_b', 'role_id_b'
        )

    def __str__(self):
        a = self.role_type_a
        if self.role_id_a:
            a += ' %s' % self.role_id_a
        b = self.role_type_b
        if self.role_id_b:
            b += ' %s' % self.role_id_b

        return '%s is part of %s' % (a, b)

    @staticmethod
    def conveyors(role_type, role_id):
        """
        Return immediate roles conveying privileges to role_type, role_id.

        Args:
            role_type (str): string describing the role type.
            role_id (str): unique ID of the role.

        Returns:
            list: list of (role_type, role_id) conveyors.
        """
        return [(rh.role_type_b, rh.role_id_b)
                for rh in RoleHierarchy.objects.filter(
                role_type_a=role_type, role_id_a=role_id)]

    @staticmethod
    def heirs(role_type, role_id):
        """
        Return immediate roles inheriting privileges from role_type, role_id.

        Args:
            role_type (str): string describing the role type.
            role_id (str): unique ID of the role.

        Returns:
            list: list of (role_type, role_id) heirs.
        """
        return [(rh.role_type_a, rh.role_id_a)
                for rh in RoleHierarchy.objects.filter(
                role_type_b=role_type, role_id_b=role_id)]

    @staticmethod
    def all_conveyors(role_type, role_id, search=BREADTH_FIRST):
        """
        Return every roles conveying privileges to role_type, role_id.

        Args:
            role_type (str): string describing the role type.
            role_id (str): unique ID of the role.
            search (int): 0: breadth-first, 1: depth-first.

        Returns:
            list: list of (role_type, role_id) conveyors.
        """
        result = []
        if search == RoleHierarchy.BREADTH_FIRST:
            line = RoleHierarchy.conveyors(role_type, role_id)
            result.extend(line)
            for r in line:
                result.extend(RoleHierarchy.all_conveyors(r[0], r[1], search))
        else:
            for r in RoleHierarchy.conveyors(role_type, role_id):
                result.append(r)
                result.extend(RoleHierarchy.all_conveyors(r[0], r[1], search))
        return result

    @staticmethod
    def all_heirs(role_type, role_id, search=BREADTH_FIRST):
        """
        Return every roles inheriting privileges from role_type, role_id.

        Args:
            role_type (str): string describing the role type.
            role_id (str): unique ID of the role.
            search (int): 0: breadth-first, 1: depth-first.

        Returns:
            list: list of (role_type, role_id) heirs.
        """
        result = []
        if search == RoleHierarchy.DEPTH_FIRST:
            for r in RoleHierarchy.heirs(role_type, role_id):
                result.append(r)
                result.extend(RoleHierarchy.all_heirs(r[0], r[1], search))
        else:
            line = RoleHierarchy.heirs(role_type, role_id)
            result.extend(line)
            for r in line:
                result.extend(RoleHierarchy.all_heirs(r[0], r[1], search))
        return result

    @staticmethod
    def get_roots():
        all_obj = RoleHierarchy.objects.all()
        b_set = set([(o.role_type_b, o.role_id_b) for o in all_obj])
        a_set = set([(o.role_type_a, o.role_id_a) for o in all_obj])
        return b_set - a_set


class RolePrivilege(models.Model):
    """Role privilege model."""

    role_type = models.CharField(_('Role type'), max_length=255)
    role_id = models.CharField(_('Role ID'), max_length=255, blank=True)

    authorized = models.BooleanField(
        _('Authorized'), default=app_settings.default_response)
    access_type = models.CharField(_('Access type'), max_length=255)

    resource_type = models.CharField(_('Resource type'), max_length=255)
    resource_id = models.CharField(_('Resource ID'), max_length=255, blank=True)  # noqa

    creation_date = models.DateTimeField(_('Created'), auto_now_add=True)
    modification_date = models.DateTimeField(_('Last modified'), auto_now=True)

    class Meta:
        """Meta class for Django."""

        verbose_name = _('Role privilege')
        verbose_name_plural = _('Role privileges')
        unique_together = (
            'role_type', 'role_id',
            'access_type',
            'resource_type', 'resource_id'
        )

    def __str__(self):
        return '%s %s %s to %s %s %s' % (
            'allow' if self.authorized else 'deny', self.role_type,
            self.role_id if self.role_id else '', self.access_type,
            self.resource_type, self.resource_id if self.resource_id else '')

    @staticmethod
    def authorize(role_type,
                  role_id,
                  perm,
                  resource_type,
                  resource_id,
                  skip_implicit=None,
                  log=None):
        """
        Authorize access to a resource to a role.

        This method checks if a role has access to a resource or a type of
        resource. Calling this method will also try to record an entry log
        in the corresponding access attempt model.

        Call will not break if there is no access attempt model. Simply,
        nothing will be recorded.

        Args:
            role_type (str): the string describing the role.
            role_id (str): the unique ID of the role.
            perm (Permission's constant): one of the permissions available
                in Permission class.
            resource_type (str): the string describing the resource.
            resource_id (str): the unique ID of the resource.
            log (bool): record an entry in access history model or not.
            skip_implicit (bool): whether to skip implicit authorization.
                It will always be skipped if you set ACCESS_CONTROL_IMPLICIT
                setting to False.

        Returns:
            bool: role has perm on resource (or not).
        """
        if skip_implicit is None:
            skip_implicit = app_settings.skip_implicit

        if log is None:
            log = app_settings.log_access

        attempt = AccessHistory(role_type=role_type, role_id=role_id,
                                resource_type=resource_type,
                                resource_id=resource_id, access_type=perm)
        attempt.response = None
        attempt.response_type = AccessHistory.EXPLICIT

        # Check role explicit perms
        attempt.response = RolePrivilege.authorize_explicit(
            role_type, role_id, perm, resource_type, resource_id)

        if attempt.response is None:

            # Else check inherited explicit perms
            for above_role_type, above_role_id in RoleHierarchy.all_conveyors(role_type, role_id):  # noqa
                attempt.response = RolePrivilege.authorize_explicit(
                    above_role_type, above_role_id, perm,
                    resource_type, resource_id)

                if attempt.response is not None:
                    attempt.conveyor_type = above_role_type
                    attempt.conveyor_id = above_role_id
                    break

        # Else check role implicit perms
        if attempt.response is None and not skip_implicit:
            attempt.response = RolePrivilege.authorize_implicit(
                role_type, role_id, perm, resource_type, resource_id)

            if attempt.response is not None:
                attempt.response_type = AccessHistory.IMPLICIT

            # Else check inherited implicit perms
            else:

                for above_role_type, above_role_id in RoleHierarchy.all_conveyors(role_type, role_id):  # noqa
                    attempt.response = RolePrivilege.authorize_implicit(
                        above_role_type, above_role_id, perm,
                        resource_type, resource_id)

                    if attempt.response is not None:
                        attempt.response_type = AccessHistory.IMPLICIT
                        attempt.conveyor_type = above_role_type
                        attempt.conveyor_id = above_role_id
                        break

        # Else give default response
        if attempt.response is None:
            attempt.response = app_settings.default_response
            attempt.response_type = AccessHistory.DEFAULT

        if log:
            attempt.save()

        return attempt.response

    @staticmethod
    def authorize_explicit(role_type,
                           role_id,
                           perm,
                           resource_type,
                           resource_id=''):
        """
        Run an explicit authorization check.

        Args:
            role_type (str): a string describing the type of role.
            role_id (str): the role's ID.
            perm (str): one of the permissions available in Permission class.
            resource_type (str): a string describing the type of resource.
            resource_id (str): the resource's ID.

        Returns:

        """
        try:
            privilege = RolePrivilege.objects.get(
                role_type=role_type, role_id=role_id,
                resource_type=resource_type, resource_id=resource_id,
                access_type=perm)
            return privilege.authorized
        except RolePrivilege.DoesNotExist:
            return None

    @staticmethod
    def authorize_implicit(role_type,
                           role_id,
                           perm,
                           resource_type,
                           resource_id=''):
        """
        Run an implicit authorization check.

        Args:
            role_type (str): a string describing the type of role.
            role_id (str): the role's ID.
            perm (str): one of the permissions available in Permission class.
            resource_type (str): a string describing the type of resource.
            resource_id (str): the resource's ID.

        Returns:
            bool: denied(perm) or allowed(perm) found in implicit_perms().
            None: if ACCESS_CONTROL_IMPLICIT is False,
                or perm is in ignored_perms.
        """
        role_instance = app_settings.mapping.get_instance(
            role_type, role_id)
        if resource_id:
            attr_name = 'can_%s_%s_%s' % (perm, resource_type, resource_id)
            if hasattr(role_instance, attr_name):
                attr = getattr(role_instance, attr_name)
                if callable(attr):
                    return attr()
                return attr
        attr_name = 'can_%s_%s' % (perm, resource_type)
        if hasattr(role_instance, attr_name):
            attr = getattr(role_instance, attr_name)
            if callable(attr):
                return attr(resource_id=resource_id)
            return attr
        attr_name = 'can_%s' % perm
        if hasattr(role_instance, attr_name):
            attr = getattr(role_instance, attr_name)
            if callable(attr):
                return attr(
                    instance=app_settings.mapping.get_instance(
                        resource_type, resource_id),
                    resource_type=resource_type, resource_id=resource_id)
            return attr
        return None

    @staticmethod
    def allow(role_type,
              role_id,
              perm,
              resource_type,
              resource_id='',
              user=None,
              log=True):
        """
        Explicitly give perm to role on resource.

        Args:
            role_type (str): a string describing the type of role.
            role_id (str): the role's ID.
            perm (str): one of the permissions available in Permission class.
            resource_type (str): a string describing the type of resource.
            resource_id (str): the resource's ID.
            user (User): an instance of settings.AUTH_USER_MODEL.
            log (bool): whether to record an entry in privileges history.

        Returns:
            access instance: the created privilege.
        """
        privilege, created = RolePrivilege.objects.update_or_create(
            role_type=role_type,
            role_id=role_id,
            access_type=perm,
            resource_type=resource_type,
            resource_id=resource_id,
            defaults={'authorized': True})

        if log:
            record = PrivilegeHistory(
                user=user, action={True: PrivilegeHistory.CREATE}.get(
                    created, PrivilegeHistory.UPDATE))
            record.update_from_privilege(privilege)

        return privilege, created

    @staticmethod
    def deny(role_type,
             role_id,
             perm,
             resource_type,
             resource_id='',
             user=None,
             log=True):
        """
        Explicitly remove perm to role on resource.

        Args:
            role_type (str): a string describing the type of role.
            role_id (str): the role's ID.
            perm (str): one of the permissions available in Permission class.
            resource_type (str): a string describing the type of resource.
            resource_id (str): the resource's ID.
            user (User): an instance of settings.AUTH_USER_MODEL.
            log (bool): whether to record an entry in privileges history.

        Returns:
            access instance: the created privilege.
        """
        privilege, created = RolePrivilege.objects.update_or_create(
            role_type=role_type,
            role_id=role_id,
            access_type=perm,
            resource_type=resource_type,
            resource_id=resource_id,
            defaults={'authorized': False})

        if log:
            record = PrivilegeHistory(
                user=user, action={True: PrivilegeHistory.CREATE}.get(
                    created, PrivilegeHistory.UPDATE))
            record.update_from_privilege(privilege)

        return privilege, created

    @staticmethod
    def forget(role_type,
               role_id,
               perm,
               resource_type,
               resource_id='',
               user=None,
               log=True):
        """
        Forget any privilege present between role and resource.

        Args:
            role_type (str): a string describing the type of role.
            role_id (str): the role's ID.
            perm (str): one of the permissions available in Permission class.
            resource_type (str): a string describing the type of resource.
            resource_id (str): the resource's ID.
            user (User): an instance of settings.AUTH_USER_MODEL.
            log (bool): whether to record an entry in privileges history.

        Returns:
            int, dict:
                the number of privileges deleted and a dictionary with the
                number of deletions per object type (django's delete return).
        """
        try:
            privilege = RolePrivilege.objects.get(
                role_type=role_type, role_id=role_id,
                resource_type=resource_type, resource_id=resource_id,
                access_type=perm)

            if log:
                record = PrivilegeHistory(
                    user=user, action=PrivilegeHistory.DELETE)
                record.update_from_privilege(privilege)

            privilege.delete()
            return True
        except RolePrivilege.DoesNotExist:
            return False


# TODO: implement HierarchyHistory model
# role creation must be recorded


class PrivilegeHistory(models.Model):
    """Privilege history model."""

    CREATE = 'c'
    # READ = 'r'  # makes no sense here
    UPDATE = 'u'
    DELETE = 'd'

    ACTIONS_VERBOSE = {
        CREATE: 'create',
        # READ: 'read',
        UPDATE: 'update',
        DELETE: 'delete'
    }

    ACTIONS = (
        (CREATE, _(ACTIONS_VERBOSE[CREATE])),
        # (READ, _(ACTIONS_VERBOSE[READ])),
        (UPDATE, _(ACTIONS_VERBOSE[UPDATE])),
        (DELETE, _(ACTIONS_VERBOSE[DELETE])),
    )

    privilege_id = models.PositiveIntegerField(
        _('Privilege reference ID'), null=True)
    reference = models.ForeignKey(
        RolePrivilege, on_delete=models.SET_NULL, null=True,
        verbose_name=_('Privilege reference'), related_name='history')
    action = models.CharField(_('Action'), max_length=1, choices=ACTIONS)
    datetime = models.DateTimeField(_('Date and time'), auto_now_add=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.SET_NULL,
        verbose_name=_('User'), related_name='privileges_changes',
        null=True)

    role_type = models.CharField(_('Role type'), max_length=255, blank=True)
    role_id = models.CharField(_('Role ID'), max_length=255, blank=True)

    authorized = models.NullBooleanField(_('Authorized'), default=None)
    access_type = models.CharField(
        _('Access type'), max_length=255, blank=True)

    resource_type = models.CharField(
        _('Resource type'), max_length=255, blank=True)
    resource_id = models.CharField(_('Resource ID'), max_length=255, blank=True)  # noqa

    class Meta:
        """Meta class for Django."""

        verbose_name = _('Privilege history')
        verbose_name_plural = _('Privilege history')

    def __str__(self):
        if self.reference:
            privilege = str(self.reference)
        else:
            privilege = RolePrivilege(
                role_type=self.role_type,
                role_id=self.role_id,
                authorized=self.authorized,
                access_type=self.access_type,
                resource_type=self.resource_type,
                resource_id=self.resource_id)

        return '[%s] user %s has %sd privilege <%d: %s>' % (
            self.datetime, self.user,
            PrivilegeHistory.ACTIONS_VERBOSE[str(self.action)],
            self.privilege_id, privilege)

    def update_from_privilege(self, privilege, save=True):
        """
        Update this object with given privilege's attributes.

        Args:
            privilege (RolePrivilege): privilege to update from.
            save (bool): whether to commit the changes to the database.
        """
        self.reference = privilege
        self.privilege_id = privilege.id
        self.role_type = privilege.role_type
        self.role_id = privilege.role_id
        self.authorized = privilege.authorized
        self.access_type = privilege.access_type
        self.resource_type = privilege.resource_type
        self.resource_id = privilege.resource_id
        if save:
            self.save()


class AccessHistory(models.Model):
    """Access history model."""

    DEFAULT = 'd'
    IMPLICIT = 'i'
    EXPLICIT = 'e'

    RESPONSE_TYPE_VERBOSE = {
        DEFAULT: 'by default',
        IMPLICIT: 'implicitly',
        EXPLICIT: 'explicitly'
    }

    RESPONSE_TYPE = (
        (DEFAULT, _(RESPONSE_TYPE_VERBOSE[DEFAULT])),
        (IMPLICIT, _(RESPONSE_TYPE_VERBOSE[IMPLICIT])),
        (EXPLICIT, _(RESPONSE_TYPE_VERBOSE[EXPLICIT]))
    )

    role_type = models.CharField(_('Role type'), max_length=255, blank=True)
    role_id = models.CharField(_('Role ID'), max_length=255, blank=True)

    # We don't want to store false info, None says "we don't know"
    response = models.NullBooleanField(_('Response'), default=None)
    response_type = models.CharField(
        _('Response type'), max_length=1, choices=RESPONSE_TYPE)
    access_type = models.CharField(_('Access'), max_length=255)

    resource_type = models.CharField(_('Resource type'), max_length=255)
    resource_id = models.CharField(_('Resource ID'), max_length=255, blank=True)  # noqa

    datetime = models.DateTimeField(_('Date and time'), default=timezone.now)

    conveyor_type = models.CharField(_('Conveyor type'), max_length=255, blank=True)  # noqa
    conveyor_id = models.CharField(_('Conveyor ID'), max_length=255, blank=True)  # noqa

    class Meta:
        """Meta class for Django."""

        verbose_name = _('Access history')
        verbose_name_plural = _('Access history')

    def __str__(self):
        inherited = ''

        if self.role_id:
            role = '%s %s' % (self.role_type, self.role_id)
        else:
            role = self.role_type

        if self.conveyor_type:
            if self.conveyor_id:
                inherited = ' (inherited from %s %s)' % (
                    self.conveyor_type, self.conveyor_id)
            else:
                inherited = self.conveyor_type

        authorized = 'authorized' if self.response else 'unauthorized'
        string = '[%s] %s was %s %s to %s %s %s' % (
            self.datetime, role,
            AccessHistory.RESPONSE_TYPE_VERBOSE[str(self.response_type)],
            authorized, self.access_type, self.resource_type, self.resource_id)
        if inherited:
            return string + inherited
        return string



# -*- coding: utf-8 -*-

"""Main test script."""

from django.test import TestCase, override_settings

import pytest
from django_fake_model import models as f

from cerberus_ac.apps import AppSettings
from cerberus_ac.models import (
    AccessHistory, PrivilegeHistory, Role, RoleHierarchy, RoleMixin,
    RolePrivilege)
from cerberus_ac.utils import get_resource_id, get_resource_type

app_settings = AppSettings()


class FakeUser(f.FakeModel, RoleMixin):
    """Fake user model."""


class FakeGroup(f.FakeModel, RoleMixin):
    """Fake group model."""


class FakeResource(f.FakeModel):
    """Fake resource model."""


@FakeUser.fake_me
@FakeGroup.fake_me
@FakeResource.fake_me
class MainTestCase(TestCase):
    """Main Django test case."""

    def setUp(self):
        """Setup method."""
        self.set_users()
        self.set_groups()
        self.set_roles()
        self.set_resources()
        self.set_role_hierarchy()
        self.set_role_privileges()

    def set_users(self):
        """Add users to self."""
        self.users = [FakeUser.objects.create() for _ in range(3)]

    def set_groups(self):
        """Add groups to self."""
        self.groups = [FakeGroup.objects.create() for _ in range(3)]

    def set_roles(self):
        """Add custom roles to self."""
        self.roles = [
            Role.objects.create(type='security'),
            Role.objects.create(type='audit', rid=15),
            Role.objects.create(type='data')
        ]

    def set_resources(self):
        """Add resources to self."""
        self.resources = [FakeResource.objects.create() for _ in range(3)]

    def test_mapping(self):
        """Test the mapping setting."""
        assert app_settings.mapping

    def test_import_classes(self):
        """Test classes imported correctly."""
        assert set(app_settings.mapping.role_classes()) == {Role, FakeUser, FakeGroup}  # noqa

    def test_getting_resource_type_and_id(self):
        """Test get_resource_type and get_resource_id methods."""
        class NoDatabaseResource(object):
            """Just an empty Python class."""

        nodb_resource_1 = NoDatabaseResource()
        nodb_resource_1.resource_id = 30
        nodb_resource_1.resource_type = 'no_db'
        nodb_resource_2 = NoDatabaseResource()

        class NoDatabaseResource2(object):
            """Fake resource class without model."""

            def resource_type(self):
                return 'user_data'

            def resource_id(self):
                return None

        nodb_resource_3 = NoDatabaseResource2()

        assert get_resource_type(nodb_resource_1) == 'no_db'
        assert get_resource_id(nodb_resource_1) == 30
        assert get_resource_type(nodb_resource_2) == 'NoDatabaseResource'
        assert get_resource_id(nodb_resource_2) is None
        assert get_resource_type(nodb_resource_3) == 'user_data'
        assert get_resource_id(nodb_resource_3) is None
        assert get_resource_type(object) == 'object'
        assert get_resource_type(1) == 'int'

    def set_role_hierarchy(self):
        """Set a role hierarchy."""
        self.users[0].take_role(self.groups[0])
        self.users[0].take_role(self.groups[1])
        self.groups[1].convey_role_to(self.users[1])
        self.roles[0].convey_role_to(self.users[0])
        self.roles[1].convey_role_to(self.users[1])
        self.roles[2].convey_role_to(self.users[2])
        self.roles[2].take_role(self.groups[2])

    def test_role_hierarchy(self):
        """Role hierarchy test method."""
        assert self.users[0].has_role(self.groups[0])
        assert self.users[0].has_role(self.groups[1])
        assert not self.users[0].has_role(self.groups[2])
        assert not self.groups[2].conveys_role_to(self.groups[1])
        assert set(self.groups[1].heirs()) == {
            self.users[0], self.users[1]}
        assert self.roles[0].conveys_role_to(self.users[0])
        assert self.roles[1].conveys_role_to(self.users[1])
        assert self.roles[2].conveys_role_to(self.users[2])
        assert set(self.users[0].conveyors()) == {
            self.groups[0], self.groups[1], self.roles[0]}
        assert set(
            self.users[2].conveyors(search=RoleHierarchy.DEPTH_FIRST)) == set(
            self.users[2].conveyors(search=RoleHierarchy.BREADTH_FIRST))
        assert set(
            self.groups[2].heirs(search=RoleHierarchy.DEPTH_FIRST)) == set(
            self.groups[2].heirs(search=RoleHierarchy.BREADTH_FIRST))

    def test_role_hierarchy_history(self):
        """Test role hierarchy history."""
        pass  # TODO: need implementation of HierarchyHistory first

    def set_role_privileges(self):
        """Set some role privileges."""
        RolePrivilege.allow('FakeUser', '1', 'do stuff on', 'FakeResource', '1')
        RolePrivilege.allow('FakeUser', '2', 'do stuff on', 'FakeResource', '2')
        RolePrivilege.allow('FakeUser', '3', 'do stuff on', 'FakeResource', '3')
        RolePrivilege.deny('FakeUser', '1', 'do stuff on', 'FakeResource', '2')
        RolePrivilege.deny('FakeUser', '2', 'do stuff on', 'FakeResource', '3')
        RolePrivilege.deny('FakeUser', '3', 'do stuff on', 'FakeResource', '1')
        RolePrivilege.allow('FakeGroup', '3', 'read', 'FakeResource', '3')
        RolePrivilege.allow('FakeGroup', '3', 'delete', 'FakeResource', '3')
        RolePrivilege.forget('FakeGroup', '3', 'delete', 'FakeResource', '3')
        RolePrivilege.forget('FakeGroup', '3', 'create', 'FakeResource', '3')
        RolePrivilege.allow('data', '', 'update', 'FakeResource', '3')

    def test_role_privileges(self):
        """Test the roles privileges."""
        # test direct explicit permission
        assert self.users[0].can('do stuff on', self.resources[0])
        assert self.users[1].can('do stuff on', self.resources[1])
        assert self.users[2].can('do stuff on', self.resources[2])
        assert not self.users[0].can('do stuff on', self.resources[1])
        assert not self.users[1].can('do stuff on', self.resources[2])
        assert not self.users[2].can('do stuff on', self.resources[0])
        assert not self.users[0].can('do stuff on', self.resources[2])
        assert not self.users[1].can('do stuff on', self.resources[0])
        assert not self.users[2].can('do stuff on', self.resources[1])
        # test indirect-1 explicit permission
        assert self.users[2].can('update', self.resources[2])
        # test indirect-2 explicit permission
        assert self.users[2].can('read', self.resources[2])
        # test direct overridden indirect-1 permission
        RolePrivilege.deny('FakeUser', '3', 'update', 'FakeResource', '3')
        assert not self.users[2].can('update', self.resources[2])
        # test direct overridden indirect-2 permission
        RolePrivilege.deny('data', '', 'read', 'FakeResource', '3')
        assert not self.users[2].can('read', self.resources[2])
        # test clashing same-level permissions

    def test_role_privileges_history(self):
        """Test role privileges history."""
        for i in (1, 2, 3):
            assert RolePrivilege.objects.get(
                role_type='FakeUser',
                role_id=i,
                access_type='do stuff on',
                authorized=True,
                resource_type='FakeResource',
                resource_id=i)
            j = {1: 2, 2: 3, 3: 1}.get(i)
            assert RolePrivilege.objects.get(
                role_type='FakeUser',
                role_id=i,
                access_type='do stuff on',
                authorized=False,
                resource_type='FakeResource',
                resource_id=j)

    def test_str_method(self):
        """Test printing methods of models."""
        self.test_role_privileges()
        for queryset in [
            Role.objects.all(),
            RolePrivilege.objects.all(),
            RoleHierarchy.objects.all(),
            AccessHistory.objects.all(),
            PrivilegeHistory.objects.all(),
            # HierarchyHistory.objects.all()[0],
        ]:
            for obj in queryset:
                print(obj)

    def tearDown(self):
        """Tear down method."""
        pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
Setup script.

Uses setuptools.
Long description is a concatenation of README.rst and CHANGELOG.rst.
"""

from __future__ import absolute_import, print_function

import io
import re
from glob import glob
from os.path import basename, dirname, join, splitext

from setuptools import find_packages, setup


def read(*names, **kwargs):
    """Read a file in current directory."""
    return io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ).read()


setup(
    name='django-cerberus-ac',
    version='0.2.4',
    license='ISC',
    description='Django Ontology Based Access Control Module',
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S)
        .sub('', read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    ),
    author='Mihnea Gheorghiu',
    author_email='mikigi92@gmail.com',
    url='https://github.com/Genida/django-cerberus-ac',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: Unix',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        # 'Framework :: Django :: 1.10',
        # 'Framework :: Django :: 1.11',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Utilities',
    ],
    keywords=[
        'django-cerberus-ac', 'access control', 'ontology-based', 'least privilege',
        'role based access control', 'django access control',
    ],
    install_requires=[
        'django-app-settings', 'django-suit-dashboard'
    ],
    entry_points={
        'archan': [
            'cerberus_ac.Privileges = cerberus_ac.plugins:Privileges'
        ]
    },
)



# -*- coding: utf-8 -*-

"""Utils module."""

from django.core.paginator import EmptyPage, PageNotAnInteger, Paginator

from .apps import AppSettings

app_settings = AppSettings()


def get_paginated_data(instances, page, num):
    paginator = Paginator(instances, num)

    try:
        paginated_data = paginator.page(page)
    except PageNotAnInteger:
        # If page is not an integer, deliver first page
        paginated_data = paginator.page(1)
    except EmptyPage:
        # If page is out of range (e.g. 9999), deliver last page of results
        paginated_data = paginator.page(paginator.num_pages)
    return paginated_data


def get_role_type(role):
    """
    Get a role's type.

    Args:
        role (obj): a Python object.

    Returns:
        str: the role's type.
    """
    if hasattr(role, 'role_type'):
        attr = role.role_type
        if callable(attr):
            return attr()
        return attr
    return app_settings.mapping.get_type(role)


def get_role_id(role):
    """
    Get a role's ID.

    Args:
        role (obj): a Python object.

    Returns:
        str: the role's ID or None.
    """
    if hasattr(role, 'role_id'):
        attr = role.role_id
        if callable(attr):
            return str(attr())
        return str(attr)
    elif hasattr(role, 'id'):
        return str(role.id)
    return ''


def get_role_type_and_id(role, role_id=''):
    """
    Return both type and ID of a role.

    Args:
        role (obj): string or role instance.
        role_id (str): string ID or None.

    Returns:
        tuple: type and ID of role.
    """
    if isinstance(role, str):
        return role, role_id
    return get_role_type(role), get_role_id(role)


def get_resource_type(resource):
    """
    Get a resource's type.

    Args:
        resource (obj): a Python object.

    Returns:
        str: the resource's type.
    """
    if hasattr(resource, 'resource_type'):
        attr = resource.resource_type
        if callable(attr):
            return attr()
        return attr
    return app_settings.mapping.get_type(resource)


def get_resource_id(resource):
    """
    Get a resource's ID.

    Args:
        resource (obj): a Python object.

    Returns:
        str: the resource's ID or None.
    """
    if hasattr(resource, 'resource_id'):
        attr = resource.resource_id
        if callable(attr):
            return attr()
        return attr
    elif hasattr(resource, 'id'):
        return resource.id
    return None


def get_resource_type_and_id(resource, resource_id=''):
    """
    Return both type and ID of a resource.

    Args:
        resource (obj): string or resource instance.
        resource_id (str): string ID or None.

    Returns:
        tuple: type and ID of resource.
    """
    if isinstance(resource, str):
        return resource, resource_id
    return get_resource_type(resource), get_resource_id(resource)



# -*- coding: utf-8 -*-

"""cerberus_ac plugins module."""

try:
    from archan import Provider, Argument, DSM as ArchanDSM

    class Privileges(Provider):
        """Dependenpy provider for Archan."""

        identifier = 'cerberus_ac.Privileges'
        name = 'Privileges'
        description = 'Provide matrix data about privileges in an access ' \
            'control scheme.'

        def get_dsm(self):
            """
            Provide matrix data about privileges in an access control scheme.

            Returns:
                archan.DSM: instance of archan DSM.
            """
            data = []
            keys = []
            return ArchanDSM(data=data, entities=keys)

except ImportError:
    class Privileges(object):
        """Empty cerberus_ac provider."""
